from pagetemplate import PageTemplate
from pagetemplate import PageTemplateFile
from pagetemplate import ViewPageTemplate
from pagetemplate import ViewPageTemplateFile

from texttemplate import TextTemplate
from texttemplate import TextTemplateFile
from texttemplate import ViewTextTemplate
from texttemplate import ViewTextTemplateFile
